#include <errno.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <ctype.h>
#include <fcntl.h>
#include <getopt.h>
#include <ncurses.h>
#include <netdb.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

#define CONFIG_FILE "/usr/local/etc/ttylink.conf"
#define BUFSIZE 32768

extern char *Version;
char rxbuffer[BUFSIZE];
char *history[100];
int rxbufp = 0, rxbufstart = 0, rxbufend = 0;
int histbufp = 0;
jmp_buf jmp_mrk;
int rx_fg_color, rx_bg_color, tx_fg_color, tx_bg_color,
    help_fg_color, help_bg_color, buffer_fg_color, buffer_bg_color,
    status_fg_color, status_bg_color, error_fg_color, error_bg_color,
    record, do_echo, bells, crmode, wmsg_showtime, wmsg_position;

typedef struct {
	char *token;
	int value;
} keyword;

enum { 
	RX_FG_COLOR, RX_BG_COLOR,
	TX_FG_COLOR, TX_BG_COLOR,
	HELP_FG_COLOR, HELP_BG_COLOR,
	BUFFER_FG_COLOR, BUFFER_BG_COLOR,
	STATUS_FG_COLOR, STATUS_BG_COLOR,
	ERROR_FG_COLOR, ERROR_BG_COLOR,
	DEFAULT_CR_MODE, DEFAULT_ECHO_MODE,
	DEFAULT_RECORD_MODE, DEFAULT_BELL_MODE,
	MSG_WINDOW_SHOW_TIME, MSG_WINDOW_POSITION
};

keyword keywords[] = {
	{"rx_fg_color",		RX_FG_COLOR},
	{"rx_bg_color",		RX_BG_COLOR},
	{"tx_fg_color",		TX_FG_COLOR},
	{"tx_bg_color",		TX_BG_COLOR},
	{"help_fg_color",	HELP_FG_COLOR},
	{"help_bg_color",	HELP_BG_COLOR},
	{"buffer_fg_color",	BUFFER_FG_COLOR},
	{"buffer_bg_color",	BUFFER_BG_COLOR},
	{"status_fg_color",	STATUS_FG_COLOR},
	{"status_bg_color",	STATUS_BG_COLOR},
	{"error_fg_color",	ERROR_FG_COLOR},
	{"error_bg_color",	ERROR_BG_COLOR},
	{"default_cr_mode",	DEFAULT_CR_MODE},
	{"default_echo_mode",	DEFAULT_ECHO_MODE},
	{"default_bell_mode",	DEFAULT_BELL_MODE},
	{"default_record_mode",	DEFAULT_RECORD_MODE},
	{"msg_window_show_time",MSG_WINDOW_SHOW_TIME},
	{"msg_window_position",	MSG_WINDOW_POSITION},
	{NULL,			0}	
};

keyword color_names[] = {
	{"black",	COLOR_BLACK},
	{"red",		COLOR_RED},
	{"green",	COLOR_GREEN},
	{"yellow",	COLOR_YELLOW},
	{"blue",	COLOR_BLUE},
	{"magenta",	COLOR_MAGENTA},
	{"cyan",	COLOR_CYAN},
	{"white",	COLOR_WHITE},
	{NULL,		0}
};

#define NONE   0
#define FLASH  1
#define SOUND  2
#define CENTER 0
#define TOP    1
#define BOTTOM 2

keyword option_names[] = {
	{"on",		TRUE},
	{"off",		FALSE},
	{"ignore",	FALSE},
	{"translate",	TRUE},
	{"none",	NONE},
	{"flash",	FLASH},
	{"sound",	SOUND},
	{"short",	100000l},
	{"medium",	300000l},
	{"long",	500000l},
	{"center",	CENTER},
	{"top",		TOP},
	{"bottom",	BOTTOM},
	{NULL,		0}
};

char *fkey_names[] =
{
	"Help",
	"Record",
	"Upload",
	"Download",
	"ViewBuffer",
	"SaveBuffer",
	"ClearBuffer",
	"EchoToggle",
	"BellsToggle",
	"CRToggle",
	"RedrawScreen",
	"Quit"
};

int ssave;

#define NORMAL_STATE 0
#define HELP_STATE   1
#define VIEW_STATE   2
#define UPLOAD_STATE 3

/*
 * Function prototypes
 */
void dohelp(WINDOW * helpwin);
char *mydate(void);
WINDOW *wmsg(attr_t attr, char *title, char *fmt,...);

int getcolor(char *color_name)
{
	int color;
	keyword *kp;

	for (color=0,kp=color_names; kp->token != NULL; kp++) {
		if (strcasecmp(kp->token,color_name) == 0) {
			color=kp->value;
			break;
		}
	}
	return color;
}

int getdefault(char *option_name)
{
	int option;
	keyword *kp;

	for (option=0,kp=option_names; kp->token != NULL; kp++) {
		if (strcasecmp(kp->token,option_name) == 0) {
			option=kp->value;
			break;
		}
	}
	return option;

}

/*
 * Read config file
 */ 
void read_config_file(void)
{
	keyword *kp;
	FILE *fp;
	char line[256], *cp1, *cp2;
	int cmd;
	char *p;

	if ((fp = fopen(CONFIG_FILE,"r")) == NULL) {
		perror(CONFIG_FILE);
		exit(2);
	} else {
		while (p = fgets(line, 256, fp),!feof(fp)) {
#ifdef DEBUG
			fprintf(stderr, "Config line read: %s",line);
#endif 
			if (line[strlen(line)-1] == '\n')
				line[strlen(line)-1] = '\0';

			if (*line == '#' || *line == '\0') continue;
			
			cp1 = strtok(line," \t=");
	 		cp2 = strtok(NULL," #=\t\n");
#ifdef DEBUG
			fprintf(stderr, "Token 1 = '%s'\n", cp1);
			fprintf(stderr, "Token 2 = '%s'\n", cp2);
#endif
			for (cmd=-1,kp=keywords; kp->token != NULL; kp++) {
				if (strcasecmp(kp->token,cp1) == 0) {
					cmd=kp->value;
					break;
				}
			}
	 		

			switch (cmd) {
				case RX_FG_COLOR:
					rx_fg_color = getcolor(cp2);
					break;
				case RX_BG_COLOR:
					rx_bg_color = getcolor(cp2);
					break;
				case TX_FG_COLOR:
					tx_fg_color = getcolor(cp2);
					break;
				case TX_BG_COLOR:
					tx_bg_color = getcolor(cp2);
					break;
				case HELP_FG_COLOR:
					help_fg_color = getcolor(cp2);
					break;
				case HELP_BG_COLOR:
					help_bg_color = getcolor(cp2);
					break;
				case BUFFER_FG_COLOR:
					buffer_fg_color = getcolor(cp2);
					break;
				case BUFFER_BG_COLOR:
					buffer_bg_color = getcolor(cp2);
					break;
				case STATUS_FG_COLOR:
					status_fg_color = getcolor(cp2);
					break;
				case STATUS_BG_COLOR:
					status_bg_color = getcolor(cp2);
					break;
				case ERROR_FG_COLOR:
					error_fg_color = getcolor(cp2);
					break;
				case ERROR_BG_COLOR:
					error_bg_color = getcolor(cp2);
					break;
				case DEFAULT_CR_MODE:
					crmode=getdefault(cp2);
					break;
				case DEFAULT_RECORD_MODE:
					record=getdefault(cp2);
					break;
				case DEFAULT_ECHO_MODE:
					do_echo=getdefault(cp2);
					break;
				case DEFAULT_BELL_MODE:
					bells=getdefault(cp2);
					break;
				case MSG_WINDOW_SHOW_TIME:
					wmsg_showtime=getdefault(cp2);
					break;
				case MSG_WINDOW_POSITION:
					wmsg_position=getdefault(cp2);
					break;
			 	default:
			 		fprintf(stderr,"ttylink.conf: illegal keyword %s\n",cp1);
					break; 
			}
		}
	}
}

/*
 * display help text in helpwin
 */
void dohelp(WINDOW * helpwin)
{
	mvwaddstr(helpwin, 0, 20, "[ Available Commands ]");
	mvwaddstr(helpwin, 15, 15, "[ Type F1 again to exit Help ]");
	mvwaddstr(helpwin, 1, 2, "F1    Show this help window, press F1 again to remove");
	mvwaddstr(helpwin, 2, 2, "F2    Record the session to file (local and remote data)");
	mvwaddstr(helpwin, 3, 2, "F3    Upload (send) a text file to remote user");
	mvwaddstr(helpwin, 4, 2, "F4    Download (receive) a text file from remote");
	mvwaddstr(helpwin, 5, 2, "F5    View scroll buffer, press F5 again to remove");
	mvwaddstr(helpwin, 6, 2, "F6    Save RX buffer to disk file (remote data only)");
	mvwaddstr(helpwin, 7, 2, "F7    Clear RX buffer");
	mvwaddstr(helpwin, 8, 2, "F8    Toggle echo on peer window (affects screen only)");
	mvwaddstr(helpwin, 9, 2, "F9    Toggle bells sounding: 0=none, 1=flash, 2=sound");
	mvwaddstr(helpwin, 10, 2, "F10   Toggle CR ignore/translate to LF mode");
	mvwaddstr(helpwin, 11, 2, "F11   Redraw screen from scratch");
	mvwaddstr(helpwin, 12, 2, "F12   Disconnect remote and quit program");
	mvwaddch(helpwin, 13, 2, ACS_UARROW);
	mvwaddstr(helpwin, 13, 3, "     Recall previous line from sent data history");
	mvwaddch(helpwin, 14, 2, ACS_DARROW);
	mvwaddstr(helpwin, 14, 3, "     Recall next line from sent data history");
	wrefresh(helpwin);
}

/*
 * Returns date&time in the DDMmmHHMMSS format
 */
char *mydate(void)
{
	time_t t;
	struct tm *tim;
	static char buf[80];

	time(&t);
	tim = gmtime(&t);
	strftime(buf, 80, "%d%b%H%M%S", tim);
	return buf;
}

/*
 * Return time in the HH:MM format
 */
char *now(void)
{
	time_t t;
	struct tm *ptm;
	static char nowstr[10];

	t = time(&t);
	ptm = gmtime(&t);
	sprintf(nowstr, "%02d:%02d", ptm->tm_hour, ptm->tm_min);
	return nowstr;
}

/*
 * Replace terminating end of line marker(s) with null
 */
char *rip(char *s)
{
	char *cp;

	cp = s + strlen(s) - 1;
	while ((cp >= s) && (*cp == '\n' || *cp == '\r'))
		*cp-- = '\0';

	return s;
}

/*
 * Return a string skipping (crmode=0) or converting (crmode!=0) CRs
 */
char *convert_CRLF(char *s, int crmode)
{
	char *cp1, *cp2;

	for (cp1 = cp2 = s; *cp1 != '\0';) {
		if (*cp1 == '\r') {
			if (crmode) {
				*cp2++ = '\n';
			}
			cp1++;
		} else {
			*cp2++ = *cp1++;
		}
	}
	*cp2 = '\0';
	return s;
}

/*
 * Return a pointer to a statically-allocated null-terminated string
 * consisting exactly of n characters; if the string has less than n
 * characters, pad it with spaces.
 */
char *strtrunc(int n, char *fmt,...)
{
#define LINELEN 256

	static char buf[LINELEN];
	int len;
	va_list ap;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	if (n > LINELEN)
		n = LINELEN;
	len = strlen(buf);
	while (len < n)
		buf[len++] = ' ';
	buf[n] = '\0';
	return buf;
}

/*
 * Ask confirmation for exit
 */
void confirmexit(void)
{
	WINDOW *w;
	char ch;
	

	w = wmsg(COLOR_PAIR(1) | A_BOLD, "Confirm exit", "Do you really want to exit program ? N\b");
	do {
		ch = toupper(wgetch(w));
	} while (ch != 'Y' && ch != 'N' && ch != '\r'); 
	if (ch == 'Y') {
		waddch(w, 'Y');
		wrefresh(w);
		shutdown(ssave, 2);
		endwin();
		exit(0);
	}
}

void ctrlc(int n)
{
	signal(SIGINT, ctrlc);
	confirmexit();
	longjmp(jmp_mrk,0);
}

/*
 * Simplest Telnet options negotiation: refuses to do any option.
 */
void doopt(int s, char *buf)
{
	unsigned char *cp1, *cp2, resp[3];

	resp[0] = IAC;

	for (cp1 = cp2 = buf; *cp1 != 0; cp1++) {
		if (*cp1 == IAC) {
			switch (*++cp1) {
				case WILL:
					resp[1] = DONT;
					break;
				case WONT:
					resp[1] = DONT;
					break;
				case DO:
					resp[1] = WONT;
					break;
				case DONT:
					resp[1] = WONT;
					break;
				case IAC:
					*cp2++ = IAC;
					continue;
			}
			resp[2] = *++cp1;
			send(s, resp, 3, 0);
		} else {
			*cp2++ = *cp1;
		}
	}
	*cp2 = '\0';
}

/*
 * Add buf to ring rx buffer
 */
void addtorxbuffer(char *buf)
{
	char *cp;
	static int over;

	for (cp = buf; *cp != '\0'; cp++) {
		rxbuffer[rxbufp++] = *cp;
		if (rxbufp > BUFSIZE - 1) {
			rxbufp = 0;
			over = 1;
		}
	}
	rxbufend = rxbufp - 1;
	if (over) {
		rxbufstart = rxbufp;
	}
}

/*
 * Save RX buffer to disk
 */
void saverxbuffer(void)
{
	char filename[80];
	int n;
	FILE *fp;
	WINDOW *wp;
	
	sprintf(filename, "TTYlink-Buffer-Save-%s", mydate());
	if ((fp = fopen(filename, "w+")) != NULL) {
		wp=wmsg(COLOR_PAIR(4), NULL, " Saving RX buffer to %s ", filename);
		for (n = rxbufstart; n != rxbufend; n++) {
			if (n > BUFSIZE - 1)
				n = 0;
			fputc(rxbuffer[n], fp);

		} 
		fclose(fp);
	} else {
		wp=wmsg(COLOR_PAIR(1), NULL, "Can't open %s: %s\n ", filename, strerror(errno));
	}

	delwin(wp);
}

/*
 * Clear RX ring buffer
 */
void clearrxbuffer(void)
{
	rxbufstart = rxbufend = rxbufp = 0;
}

/*
 * Count lines in rx buffer
 */
int countlines(void)
{
	char *cp;
	int n = 0;

	for (cp = &rxbuffer[rxbufstart]; cp != &rxbuffer[rxbufend + 1]; cp++) {
		if (*cp == '\n') {
			n++;
		}
	}
	return n;
}

/*
 * Display nlines lines of RX buffer beginning at line start and column pos
 * in the window rxwin
 */
void disprxbuffer(WINDOW * rxwin, int start, unsigned int pos)
{
	int n, i;
	char *cp, tmp[1024];

	werase(rxwin);

	/*
	 * Skip start lines
	 */
	for (n = 0, cp = &rxbuffer[rxbufstart]; n < start && cp != &rxbuffer[rxbufend + 1]; cp++) {
		if (cp > &rxbuffer[BUFSIZE - 1]) {
			cp = rxbuffer;
		}
		if (*cp == '\n') {
			n++;
		}
	}
	for (n = 0, i = 0; n < 18 && cp != &rxbuffer[rxbufend + 1]; cp++) {
		if (cp > &rxbuffer[BUFSIZE - 1]) {
			cp = rxbuffer;
		}
		if (*cp == '\n') {
			tmp[i] = '\0';
			if (strlen(tmp) > pos) {
				mvwaddstr(rxwin, n, 0, strtrunc(70, "%s", &tmp[pos]));
			}
			n++;
			i = 0;
			continue;
		}
		tmp[i++] = *cp;
	}
	wrefresh(rxwin);
}

/*
 * Add a line to history buffer
 */
int addtohistory(char *buf)
{
	if (histbufp > 98) {
		histbufp = 98;
		free(history[0]);
		memmove(&history[0], &history[1], 98*sizeof(char *));
	}
	history[histbufp++] = strdup(rip(buf));
	history[histbufp] = "";
	return histbufp;
}

/*
 * Display a query, message or error window
 */
WINDOW *wmsg(attr_t attr, char *title, char *fmt,...)
{
	WINDOW *msgwin;
	va_list ap;
	char buf[256], tbuf[256];
	int len;
	struct timeval now, then;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	len = strlen(buf);
	switch(wmsg_position) {
		case TOP:
			msgwin = newwin(3, len + 2, 1, COLS / 2 - len / 2);
			break;
		case BOTTOM:
			msgwin = newwin(3, len + 2, LINES - 4 , COLS / 2 - len / 2);
			break;
		case CENTER:
		default:
			msgwin = newwin(3, len + 2, LINES / 2 - 1, COLS / 2 - len / 2);
			break;
	}
	wattrset(msgwin, attr);
	wbkgd(msgwin, attr);
	box(msgwin, 0, 0);

	if (title == NULL)
		title = "Message";
	sprintf(tbuf, "[ %s ]", title);
	mvwprintw(msgwin, 0, (len - strlen(tbuf) + 2) / 2, tbuf);
	mvwprintw(msgwin, 1, 1, "%s", buf);
	wrefresh(msgwin);
	gettimeofday(&then,NULL);
	do { 
		gettimeofday(&now,NULL);
	} while (((now.tv_sec-then.tv_sec)*1000000l+now.tv_usec-then.tv_usec) < wmsg_showtime);
	return msgwin;
}

/*
 * Manage incoming bells doing nothing, flash or beep
 */
void dobells(char *buf, int bells_active)
{
	switch (bells_active) {
		case NONE:
			return;
		case FLASH:
			if (strchr(buf, '\a') != NULL)
				flash();
			break;
		case SOUND:
			if (strchr(buf, '\a') != NULL)
				beep();
			break;
	}
}

/*
 * Show upload, download, etc. status and time
 */
void show_status(WINDOW * w, int record, int upload, int download, int do_echo, int bells, int crmode)
{
	char buf[80];

	sprintf(buf, "[%c%c%c%c%c%c]",
		(record) ? 'R' : '-',
		(upload) ? 'U' : '-',
		(download) ? 'D' : '-',
		(do_echo) ? 'E' : '-',
		(bells) ? (bells == FLASH) ? 'F' : 'S' : 'I',
		(crmode) ? 'T' : 'I');
	wattrset(w, COLOR_PAIR(6) | A_BOLD);
	mvwaddstr(w, 0, 4, buf);
	sprintf(buf, "[%s]", now());
	mvwaddstr(w, 0, COLS - 11, buf);
	wrefresh(w);
}

/*
 * Main program
 */
int main(int argc, char *argv[])
{
	WINDOW *parentwin = NULL, *opeerwin = NULL, *omywin = NULL, *peerwin = NULL, *mywin = NULL, *helpwin = NULL, *oviewwin = NULL,
	      *viewwin = NULL, *tmpwin = NULL;
	char *hostname = NULL;
	char buf[1024], line[1024], ttyname[10], *cp;
	fd_set fset;
	int port = IPPORT_TTYLINK;
	int len, res, ch, n, newx, x, y, size, diff, margin, excess;
	int s, f1, start, nlines, totlines, hbp, state, upload;
	unsigned int pos;
	int download, topty;
	struct hostent *host;
	struct in_addr addr;
	struct sockaddr_in fsocket;
	struct timeval tval;
	struct termios trm;
	FILE *rfp = NULL, *dfp = NULL, *ufp = NULL;

	read_config_file();

	topty = (strstr(argv[0], "ttydriv") != NULL);

	if (topty) {
		if (argc != 2) {
			printf("\nUsage: %s <pty name>   (/dev/ttyp prefix can be omitted)\n\n", argv[0]);
			return 1;
		}
	} else {
		if ((argc < 2) || (argc > 3)) {
			printf("\nUsage: %s <hostname> [<port>]   (default port is TTYLINK)\n\n", argv[0]);
			return 1;
		}
		if (argc == 3) {
			port = atoi(argv[2]);
		}
	}

	parentwin = initscr();
	cbreak();
	noecho();
	nonl();

	if (has_colors()) {
		start_color();
		init_pair(1, error_fg_color, error_bg_color);	/* error window	*/
		init_pair(2, tx_fg_color, tx_bg_color);		/* tx window 	*/
		init_pair(3, rx_fg_color, rx_bg_color);		/* rx window 	*/
		init_pair(4, help_fg_color, help_bg_color);	/* help 	*/
		init_pair(5, buffer_fg_color, buffer_bg_color);	/* rx buffer	*/
		init_pair(6, status_fg_color, status_bg_color);	/* status line	*/
	} else {
		/* only white and black */
		init_pair(1, COLOR_BLACK, COLOR_WHITE);
		init_pair(2, COLOR_WHITE, COLOR_BLACK);
		init_pair(3, COLOR_WHITE, COLOR_BLACK);
		init_pair(4, COLOR_WHITE, COLOR_BLACK);
		init_pair(5, COLOR_WHITE, COLOR_BLACK);
		init_pair(6, COLOR_WHITE, COLOR_BLACK);
	}
	werase(parentwin);
	wattrset(parentwin, COLOR_PAIR(6) | A_BOLD);
	box(parentwin, 0, 0);
	mvwprintw(parentwin, 0, COLS / 2 - strlen(Version) / 2 - 2, "[ %s ]", Version);
	wrefresh(parentwin);

	omywin = subwin(parentwin, (LINES % 2) ? LINES / 4 : LINES / 4 - 1, COLS - 2, 3 * LINES / 4, 1);
	wattrset(omywin, COLOR_PAIR(2));
	werase(omywin);
	box(omywin, 0, 0);
	mywin = subwin(omywin, (LINES % 2) ? LINES / 4 - 2 : LINES / 4 - 3, COLS - 4, 3 * LINES / 4 + 1, 2);
	wattrset(mywin, COLOR_PAIR(2));
	wbkgd(mywin, COLOR_PAIR(2));
	scrollok(mywin, TRUE);
	nodelay(mywin, TRUE);
	keypad(mywin, TRUE);

	opeerwin = subwin(parentwin, 3 * LINES / 4 - 1, COLS - 2, 1, 1);
	wattrset(opeerwin, COLOR_PAIR(3));
	werase(opeerwin);
	box(opeerwin, 0, 0);
	peerwin = subwin(opeerwin, 3 * LINES / 4 - 3, COLS - 4, 2, 2);
	wattrset(peerwin, COLOR_PAIR(3));
	wbkgd(peerwin, COLOR_PAIR(3));
	scrollok(peerwin, TRUE);

	/*
	 * Print function key labels on bottom line
	 */
	wattrset(parentwin, COLOR_PAIR(6) | A_REVERSE);

	size = (COLS - 2) / 12;
	diff = (COLS - 12 * size);
	margin = diff / 2;
	excess = 0;
	if (diff > 4) {
		/* we have space for 3 more characters on labels 10 11 & 12 */
		margin = (diff - 3) / 2;
		excess = 1;
	}
	for (n = 0; n < 12; n++) {
		mvwprintw(parentwin, LINES - 1, n * size + margin + excess, "%d%s", n + 1, strtrunc(size - (n > 8 ? (excess ? 2 : 3) : 2), "%s", fkey_names[n]));
		if (excess && n > 8)
			excess++;
	}
	wrefresh(parentwin);
	wrefresh(opeerwin);
	wrefresh(peerwin);
	wrefresh(omywin);
	wrefresh(mywin);

	signal(SIGINT, ctrlc);

	if (topty) {
		if (strstr(argv[1], "/dev/") == NULL) {
			if (isdigit(*argv[1])) {
				/* only tty number N given */
				sprintf(buf, "/dev/ttyp%d", atoi(argv[1]));
			} else {
				/* only tty name given: ttypN */
				sprintf(buf, "/dev/%s", argv[1]);
			}
		} else {
			/* full device name given: /dev/ttypN */
			strcpy(buf, argv[1]);
		}
		strcpy(ttyname, &buf[5]);

		if ((f1 = open(buf, O_RDWR)) == -1) {
			delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Unable to open %s", buf));
			exit(1);
		}
		tcgetattr(f1, &trm);
		trm.c_iflag &= ~IGNCR;
		trm.c_iflag &= ~ICRNL;
		tcsetattr(f1, TCSANOW, &trm);

		/*
		 * How to get the name of the peer connected to other side of the pty ?
		 */
		hostname = "Remote user";
		sprintf(buf, "[ %s ]", hostname);
		mvwaddstr(opeerwin, 0, (COLS - strlen(buf)) / 2, buf);
		wrefresh(parentwin);
		res = write(f1, " \b", 2);
		fflush(stdout);
	} else {
		hostname = strdup(argv[1]);

		delwin(wmsg(COLOR_PAIR(4), NULL, "Resolving %s... ", hostname));
		if ((host = gethostbyname(hostname)) == 0) {
			delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Unknown host name %s", hostname));
			endwin();
			return 1;
		}
		memset(&fsocket, 0, sizeof(fsocket));
		addr.s_addr = *(long *) host->h_addr_list[0];
		fsocket.sin_family = AF_INET;
		fsocket.sin_port = htons(port);
		fsocket.sin_addr = addr;

		wrefresh(peerwin);
		delwin(wmsg(COLOR_PAIR(4), NULL, "Trying %u.%u.%u.%u:%u...",
			    (unsigned char) host->h_addr_list[0][0],
			    (unsigned char) host->h_addr_list[0][1],
			    (unsigned char) host->h_addr_list[0][2],
			    (unsigned char) host->h_addr_list[0][3],
			    ntohs(fsocket.sin_port)));

		if ((s = ssave = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
			delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Can't create socket"));
			endwin();
			return 1;
		}
		if (connect(s, (struct sockaddr *) &fsocket, sizeof(fsocket)) == -1) {
			delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Connect failed"));
			endwin();
			return 1;
		}
		wrefresh(peerwin);
		delwin(wmsg(COLOR_PAIR(4), NULL, "Connected to %s", hostname));
		sprintf(buf, "[ %s ]", hostname);
		mvwaddstr(opeerwin, 0, (COLS - strlen(buf)) / 2, buf);
		f1 = s;
	}
	sprintf(buf, "[ %s ]", getenv("LOGNAME"));
	mvwaddstr(omywin, 0, (COLS - strlen(buf)) / 2, buf);
	wrefresh(parentwin);
	wrefresh(omywin);

	n = 0;
	hbp = 0;
	pos = 0;
	start = 0;
	nlines = 18;
	totlines = 0;
	history[0] = "";
	state = NORMAL_STATE;
	download = upload = 0;

	if (record) {
		sprintf(buf, "TTYlink-Record-%s", mydate());
		if ((rfp = fopen(buf, "wb+")) == NULL) {
			delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Cant' open record filename"));
		}
	}

	setjmp(jmp_mrk);
	while (1) {
		switch (state) {
			case NORMAL_STATE:
				show_status(parentwin, record, upload, download, do_echo, bells, crmode);
				wprintw(mywin, " \b");
				wrefresh(mywin);
				break;
			case HELP_STATE:
				wrefresh(helpwin);
				break;
			case VIEW_STATE:
				wrefresh(oviewwin);
				disprxbuffer(viewwin, start, pos);
				break;
		}

		FD_ZERO(&fset);
		FD_SET(STDIN_FILENO, &fset);
		FD_SET(f1, &fset);

		tval.tv_sec = 60;
		tval.tv_usec = 0;

		res = select(f1 + 1, &fset, NULL, NULL, &tval);

		if (res == 0) {
			/* timeout */
			continue;
		}
		if (FD_ISSET(f1, &fset)) {
			len = read(f1, buf, sizeof(buf));
			if (len > 0) {
				buf[len] = '\0';
				doopt(f1, buf);
				convert_CRLF(buf,crmode);
				dobells(buf, bells);
				waddstr(peerwin, buf);
				wrefresh(peerwin);
				if (record) {
					fwrite(buf, 1, strlen(buf), rfp);
				}
				if (download) {
					fwrite(buf, 1, strlen(buf), dfp);
				}
				addtorxbuffer(buf);
				wprintw(mywin, " \b");
				wrefresh(mywin);
			} else {
				wmsg(COLOR_PAIR(4), NULL, "Disconnected from remote host");
				endwin();
				exit(0);
			}
		}
		if (FD_ISSET(STDIN_FILENO, &fset)) {
			ch = wgetch(mywin);
			switch (ch) {
				case '\r':
					waddch(mywin, '\n');
					wrefresh(mywin);
					line[n++] = '\n';
					line[n] = '\0';
					res = write(f1, line, n);
					if (record) {
						fwrite(line, 1, n, rfp);
					}
					if (do_echo) {
						wattrset(peerwin, COLOR_PAIR(3) | A_BOLD);
						wprintw(peerwin, "%s", line);
						wrefresh(peerwin);
						wattrset(peerwin, COLOR_PAIR(3));
					}
					hbp = addtohistory(line);
					n = 0;
					break;
				case KEY_BACKSPACE:
				case 127:
					if (n > 0) {
						n--;
						waddch(mywin, '\b');
						wdelch(mywin);
						wrefresh(mywin);
					}
					break;
				case 9:
					do {
						waddch(mywin, ' ');
						line[n++] = ' ';
					} while (n % 8);
					wrefresh(mywin);
					break;
				case KEY_F(1):
					if (state == NORMAL_STATE) {
						state = HELP_STATE;
						helpwin = newwin(16, 60, LINES / 2 - 7, COLS / 2 - 30);
						wattrset(helpwin, COLOR_PAIR(4));
						wbkgd(helpwin, COLOR_PAIR(4));
						box(helpwin, 0, 0);
						wrefresh(helpwin);
						dohelp(helpwin);
						break;
					}
					if (state == HELP_STATE) {
						delwin(helpwin);
						touchwin(opeerwin);
						touchwin(omywin);
						touchwin(peerwin);
						touchwin(mywin);
						wrefresh(opeerwin);
						wrefresh(omywin);
						wrefresh(peerwin);
						wrefresh(mywin);
						state = NORMAL_STATE;
						break;
					}
					break;
				case KEY_F(2):
					record = !record;
					if (record) {
						sprintf(buf, "TTYlink-Record-%s", mydate());
						if ((rfp = fopen(buf, "wb+")) == NULL) {
							delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Cant' open record filename"));
							record = 0;
						}
					} else {
						fclose(rfp);
					}
					break;
				case KEY_F(3):
					upload = TRUE;
					state = UPLOAD_STATE;
					show_status(parentwin, record, upload, download, do_echo, bells, crmode);
					tmpwin = wmsg(COLOR_PAIR(4), "Enter filename to upload:", "%66s", "");
					echo();
					wmove(tmpwin, 1, 1);
					wgetnstr(tmpwin, line, 64);
					noecho();
					delwin(tmpwin);
					if ((ufp = fopen(line, "r")) == NULL) {
						delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "%s", strerror(errno)));
					} else {
						while ((len = fread(buf, 1, strlen(buf), ufp)) > 0) {
							res = write(f1, buf, len);
						}
						fclose(ufp);
					}
					upload = FALSE;
					state = NORMAL_STATE;
					break;
				case KEY_F(4):
					download = !download;
					if (download) {
						sprintf(buf, "TTYlink-Download-%s", mydate());
						if ((dfp = fopen(buf, "wb+")) == NULL) {
							delwin(wmsg(COLOR_PAIR(1) | A_BOLD, "ERROR", "Cant' open download filename"));
						}
					} else {
						fclose(dfp);
					}
					break;
				case KEY_F(5):
					if (state == NORMAL_STATE) {
						state = VIEW_STATE;
						start = 0;
						pos = 0;
						totlines = countlines();
						nlines = (totlines > 4 * LINES / 5 - 2) ? 4 * LINES / 5 - 2 : totlines;
						oviewwin = newwin(4 * LINES / 5, 9 * COLS / 10, LINES / 10, COLS / 20);
						wattrset(oviewwin, COLOR_PAIR(5));
						wbkgd(oviewwin, COLOR_PAIR(5));
						box(oviewwin, 0, 0);
						mvwaddstr(oviewwin, 0, 9 * COLS / 20 - 7, "[ RX Buffer ]");
						mvwaddstr(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 19, "[    , Pg , Pg , Home, End, F5=exit]");
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 18, ACS_UARROW);
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 17, ACS_DARROW);
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 16, ACS_RARROW);
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 15, ACS_LARROW);
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 10, ACS_UARROW);
						mvwaddch(oviewwin, 4 * LINES / 5 - 1, 9 * COLS / 20 - 5, ACS_DARROW);
						wrefresh(oviewwin);
						viewwin = derwin(oviewwin, 4 * LINES / 5 - 2, 9 * COLS / 10 - 2, 1, 1);
						wattrset(viewwin, COLOR_PAIR(5));
						wbkgd(viewwin, COLOR_PAIR(5));
						keypad(viewwin, TRUE);
						scrollok(viewwin, TRUE);
						break;
					}
					if (state == VIEW_STATE) {
						delwin(viewwin);
						delwin(oviewwin);
						touchwin(opeerwin);
						touchwin(omywin);
						touchwin(peerwin);
						touchwin(mywin);
						wrefresh(opeerwin);
						wrefresh(omywin);
						wrefresh(peerwin);
						wrefresh(mywin);
						state = NORMAL_STATE;
						break;
					}
					break;
				case KEY_F(6):
					saverxbuffer();
					break;
				case KEY_F(7):
					clearrxbuffer();
					break;
				case KEY_F(8):
					do_echo = ~do_echo;
					break;
				case KEY_F(9):
					bells = (bells + 1) % 3;
					break;
				case KEY_F(10):
					crmode = !crmode;
					break;
				case KEY_F(11):
					touchwin(opeerwin);
					touchwin(omywin);
					touchwin(peerwin);
					touchwin(mywin);
					wrefresh(peerwin);
					wrefresh(mywin);
					wrefresh(curscr);
					break;
				case KEY_F(12):
					confirmexit();
					break;
				case KEY_UP:
					if (state == VIEW_STATE) {
						start = (start > 1) ? start - 1 : 0;
					}
					if (state == NORMAL_STATE && histbufp > 0) {
						hbp = (hbp <= 0) ? 0 : hbp - 1;
						strcpy(line, history[hbp]);
						wprintw(mywin, "\r%s", line);
						wclrtoeol(mywin);
						wrefresh(mywin);
						n = strlen(line);
					}
					break;
				case KEY_PPAGE:
					if (state != VIEW_STATE)
						break;
					start = (start > nlines) ? start - nlines : 0;
					break;
				case KEY_DOWN:
					if (state == VIEW_STATE) {
						start = (start + 1 <= totlines - nlines) ? start + 1 : start;
					}
					if (state == NORMAL_STATE && histbufp > 0) {
						hbp = (hbp + 1 < histbufp) ? hbp + 1 : histbufp;
						strcpy(line, history[hbp]);
						wprintw(mywin, "\r%s", line);
						wclrtoeol(mywin);
						wrefresh(mywin);
						n = strlen(line);
					}
					break;
				case KEY_NPAGE:
					if (state != VIEW_STATE)
						break;
					start = (start + nlines < totlines - nlines) ? start + nlines : (totlines >= nlines) ? totlines - nlines : 0;
					break;
				case KEY_LEFT:
					if (state != VIEW_STATE)
						break;
					pos = (pos > 0) ? pos - 1 : 0;
					break;
				case KEY_RIGHT:
					if (state != VIEW_STATE)
						break;
					pos++;
					break;
				case KEY_HOME:
					if (state != VIEW_STATE)
						break;
					start = 0;
					break;
				case KEY_END:
					if (state != VIEW_STATE)
						break;
					start = totlines - nlines;
					break;
				case KEY_DC:
					break;
				default:
					if (state != NORMAL_STATE)
						break;
					if (isprint(ch) || ch > 127) {
						waddch(mywin, ch);
					} else {
						wattrset(mywin, COLOR_PAIR(2) | A_REVERSE | A_BOLD);
						waddch(mywin, ch+'@');
						wattrset(mywin, COLOR_PAIR(2));
					}
					waddstr(mywin, " \b");
					wrefresh(mywin);
					line[n++] = (char) ch;
					if (n > 74) {
						line[n] = '\0';
						rip(line);
						cp = strrchr(line, ' ');
						if (cp != NULL) {
							/* we can safely owerwrite the blank */
							newx = cp - line;
						} else {
							/* we must create a blank place */
							newx = 74;
							memcpy(&line[newx + 1], &line[newx], 3);
						}
						strcpy(buf, line);
						buf[newx] = '\n';
						buf[newx + 1] = '\0';
						res = write(f1, buf, newx + 1);
						if (record) {
							fwrite(buf, 1, newx + 1, rfp);
						}
						if (do_echo) {
							wattrset(peerwin, COLOR_PAIR(3) | A_BOLD);
							wprintw(peerwin, "%s", buf);
							wrefresh(peerwin);
							wattrset(peerwin, COLOR_PAIR(3));
						}
						hbp = addtohistory(buf);
						strcpy(line, &line[newx + 1]);
						n = strlen(line);
						getyx(mywin, y, x);
						wmove(mywin, y, newx);
						wclrtoeol(mywin);
						waddstr(mywin, "\n");
						waddstr(mywin, line);
						wrefresh(mywin);
					}
					break;
			}
			continue;
		}
	}
	endwin();
	exit(0);
}
